/*
** Module   : ToggleBox.java
** Abstract : toggle box implementation
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20140610 Initial implementation.
*/

package com.goldencode.p2j.ui.client;

import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * An implementation of a "toggle box" with a text label.
 */
public abstract class ToggleBox
extends AbstractWidget
implements KeyListener,
           FocusListener, 
           DataContainer
{
   /** Toggle Box label. */
   private String textLabel;

   /** Toggle Box attribute container. */
   private ToggleBoxConfig config = null;

   /** Widget in focus flag. */
   protected boolean widgetFocused = false;
   
   /** Previous cursor visibility flag to restore on focus lost. */
   private boolean curvis = false;
   
   /**
    * This constructor creates a toggle-box with a given text label.
    * 
    * @param   owner
    *          The parent container, frame or dialog which will include this toggle-box.
    * @param   text
    *          The string to be used as the toggle-box text.
    */
   public ToggleBox(Container owner, String text)
   {
      setTextLabel(text);
      
      config = new ToggleBoxConfig(ComponentConfig.TOGGLE_BOX, this);
      
      config.setName(text);
      config.setWidthCharsDirect(width());
      config.setHelp(null);
      addFocusListener(this);
   }

   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   cfg
    *          Configuration data.
    */
   public ToggleBox(ComponentConfig cfg)
   {
      setTextLabel("");
      config = new ToggleBoxConfig(this, cfg);
      config.applyConfig(cfg);

      if (cfg.getLabel() != null)
         setText(cfg.getLabel());
      addFocusListener(this);
   }

   /**
    * Get widget dimension.
    * 
    * @return  widget dimension.
    */
   @Override
   public Dimension dimension()
   {
      return new Dimension(width(), height());
   }

   /**
    * Get label text.
    * 
    * @return  label text.
    */
   public String textLabel()
   {
      return textLabel;
   }

   /**
    * Provide access to the attributes stored in the widget.
    * 
    * @return  Reference to attribute container.
    */
   @Override
   @SuppressWarnings("unchecked")
   public ToggleBoxConfig config()
   {
      return config;
   }

   /**
    * Draw the button. Called by this ToggleBox's parent container.
    */
   @Override
   public abstract void draw();

   // DataContainer methods implementations start
   /**
    * Get label text.
    * 
    * @return  label text.
    */
   @Override
   public String getText()
   {
      return textLabel;
   }

   /**
    * Getting the BaseDataType object representing the current value.
    *
    * @return  The value representing the current widget value.
    */
   @Override
   public BaseDataType getValue()
   {
      return new logical(config.isChecked());
   }

   /**
    * Getting the modification state of the widget.
    * 
    * @return  <code>true</code> if the widget value was modified,
    *          <code>false</code> otherwise.
    */
   @Override
   public boolean isModified()
   {
      return config.isModified();
   }

   /**
    * Sets the value of the toggle-box (the checked state) based on a
    * BaseDataType value.
    * 
    * @param    value
    *           The BaseDataType value of the logical value for toggle-box.
    */
   @Override
   public void setValue(BaseDataType value)
   {
      if (value instanceof logical)
      {
         config.setChecked(((logical)value).booleanValue());
      }
   }
   // DataContainer methods implementations end
   
   /**
    * Calculate the height of this widget.
    *
    * @return    The height in rows.
    */
   @Override
   public int height()
   {
      return config.getHeightChars() > 0 ? config.getHeightChars() : 1;
   }
   
   /**
    * Calculate the width of this widget.
    *
    * @return    The width in characters.
    */
   @Override
   public int width()
   {
      return config.getWidthChars() > 0 ? config.getWidthChars() : textLabel().length() + 3;
   }
   
   /**
    * Set the new text for the component.
    * 
    * @param   text
    *          New text value.
    */
   public void setText(String text)
   {
      setTextLabel(text);

      repaint();
      
      if (config.isAutoResize())
         config.setWidthCharsDirect(textLabel().length() + 3);

   }

   /**
    * Set toggle-box label.
    * 
    * @param   label
    *          Label text.
    */
   public void setTextLabel(String label)
   {
      this.textLabel = label;
   }

   /**
    * Get widget minimum size.
    * 
    * @return  widget minimum size.
    */
   @Override
   public Dimension minimumSize()
   {
      return dimension();
   }

   /**
    * This is called when the toggle-box takes the input focus. When the help string is defined
    * for the toggle-box it should be displayed at the window status line.
    *
    * @param   event
    *          The focus event to be processed.
    */
   @Override
   public void onFocusGained(FocusEvent event)
   {
      // no need in code below if widget already focused
      if (widgetFocused)
      {
         return;
      }
      
      // force the cursor off since it should not display while we have focus
      curvis = screen().setCursorStatus(false);

      // mark widget as gained input focus      
      widgetFocused = true;
      
      // even for consecutive FOCUS_GAINED events, the HELP text must be set
      // for this toggle-box, because the HELP message-text may have been changed
      // for another widget (i.e. on a nested UPDATE)
      
      // draw the help string for focused toggle-box
      ThinClient.setHelp(config.getHelp());
   }
   
   /**
    * This is called when the button looses the input focus.
    *
    * @param   focusEvent
    *          The focus event to be processed.
    */
   @Override
   public void onFocusLost(FocusEvent focusEvent)
   {
      // no need in code below if widget already lost the focus
      if (!widgetFocused)
      {
         return;
      }
      
      widgetFocused = false;
      
      // restore the cursor visibility that existed when we gained focus 
      screen().setCursorStatus(curvis);
   }

   /**
    * Check if widget needs in cusror to be on when having the input focus.
    * 
    * @return  <code>true</code> if cursor is required.
    */
   @Override
   public boolean forceCursorOnFocus()
   {
      return false;
   }

   /**
    * This method is invoked when key is pressed.
    * 
    * @param   event
    *          Event which triggered this call.
    */
   @Override
   public void onKeyPressed(KeyInput event)
   {
   }

   /**
    * This method is invoked when key is typed.
    * 
    * @param   event
    *          Event which triggered this call.
    */
   @Override
   public void onKeyTyped(KeyInput event)
   {
   }
   
   /**
    * Processes key events occurring on this object by dispatching them
    * to any registered KeyListener objects.
    *
    * @param    keyEvent
    *           The event to be handled.
    */
   @Override
   public void processKeyEvent(KeyInput keyEvent)
   {
      // use arrow keys to move focus between widgets (only left/right work
      // in the alert box case)
      if (parent() instanceof AlertBox)
      {
         setActionCode(keyEvent);
      }
      else
      {
         if (handleCursorKeys(keyEvent.actionCode(), (Frame) UiUtils.locateFrame(this)))
            return;
      }

      handleBasicEvents(keyEvent);
   }

   /**
    * Handle cursor movement keys.
    * 
    * @param   key
    *          Key code.
    * @param   frame
    *          Parent frame.
    * 
    * @return  <code>true</code> if any cursor key was found.
    */
   private boolean handleCursorKeys(int key, Frame frame)
   {
      boolean returnNow = false;
      
      if (key == Keyboard.KA_TAB)
      {
         returnNow = nextFocus();
      }
      else if (key == Keyboard.KA_BACK_TAB)
      {
         returnNow = prevFocus();
      }
      else if (key == Keyboard.KA_CURSOR_LEFT)
      {
         // the left most widget ignores this condition
         if (!frame.isLeftMostWidget(this))
         {
            returnNow = prevFocus();
         }
      }
      else if (key == Keyboard.KA_CURSOR_UP)
      {
         returnNow = moveAboveWidget(frame);
      }
      else if (key == Keyboard.KA_CURSOR_RIGHT)
      {
         // the right most widget ignores this condition
         if (!frame.isRightMostWidget(this))
         {
            returnNow = nextFocus();
         }
      }
      else if (key == Keyboard.KA_CURSOR_DOWN)
      {
         returnNow = moveBelowWidget(frame);
      }
      return returnNow;
   }

   /**
    * Move focus to widget below in specified frame.
    * 
    * @param   frame
    *          Parent frame.
    *          
    * @return  <code>true</code>
    */
   private boolean moveBelowWidget(Frame frame)
   {
      frame.moveBelowWidget(this, 0);
      refresh();
      return true;
   }

   /**
    * Move focus to widget above in specified frame.
    * 
    * @param   frame
    *          Parent frame.
    *          
    * @return  <code>true</code>
    */
   private boolean moveAboveWidget(Frame frame)
   {
      frame.moveAboveWidget(this, 0);
      refresh();
      return true;
   }

   /**
    * Move focus to previous widget in Tab order.
    * 
    * @return  <code>true</code>
    */
   private boolean prevFocus()
   {
      ThinClient.previousFocus(this);
      refresh();
      return true;
   }

   /**
    * Move focus to next widget in Tab order.
    * 
    * @return  <code>true</code>
    */
   private boolean nextFocus()
   {
      ThinClient.nextFocus(this);
      refresh();
      return true;
   }

   /**
    * Refresh widget at the screen.
    */
   private void refresh()
   {
      draw();
      requestSync();
   }

   /**
    * Prepare event for processing.
    *   
    * @param   ki
    *          Input event.
    */
   private void setActionCode(KeyInput ki)
   {
      int key = ki.actionCode();
      // alert box processing has no need to generate entry/leave events 
      // so we take a very simple approach
      if (key == Keyboard.KA_CURSOR_LEFT)
      {
         ki.setActionCode(Keyboard.KA_BACK_TAB);
         ki.setKeyCode(Key.VK_BACK_TAB);
      }
      else if (key == Keyboard.KA_CURSOR_RIGHT)
      {
         ki.setActionCode(Keyboard.KA_TAB);
         ki.setKeyCode(Key.VK_TAB);         
      }
   }
   
   /**
    * Process basic events - TAB, BACK-TAB and ENTER key event.
    * 
    * @param   event
    *          Key event to process.
    */
   private void handleBasicEvents(KeyInput event)
   {
      super.processKeyEvent(event);

      if (event.isConsumed())
      {
         return;
      }

      int key = event.keyCode();
      if (key == '\t')
      {
         parent().nextFocus();
         return;
      }
      else if (key == Key.VK_BACK_TAB)
      {
         parent().prevFocus();
         return;
      }

      if (super.isEnabled() && (key == Key.VK_ENTER || key == ' '))
      {
         // inverse the CHECKED state
         config.setChecked(!config.isChecked());
         config.setModified(true);
         refresh();
         // Generate VALUE_CHANGED event
         KeyInput valueChangedEvent = new KeyInput(this, EventType.KEY_TYPED,
                                                   Keyboard.SE_VALUE_CHANGED);
         EventManager.postEvent(valueChangedEvent);
         // eat the original event
         event.consume();
      }
   }
}
